import pathlib

from spoonbill.i18n import extract_schema_po
from tests.data import awards_columns, contracts_columns, parties_columns, planning_columns, tenders_columns

here = pathlib.Path(__file__).parent


def test_babel_extractor():
    with open(here / "data" / "ocds-simplified-schema.json", "rb") as schema:
        result = [msg for msg in extract_schema_po(schema, "", "", {"combine": "documents,amendments,milestones"})]
    msg_ids = [c[2] for c in result]
    for columns in (tenders_columns, parties_columns, contracts_columns, awards_columns, planning_columns):
        for col in columns:
            assert col in msg_ids



import gettext
import json
import locale
from collections import deque
from functools import lru_cache

import jsonref
import pkg_resources

from spoonbill.common import COMBINED_TABLES
from spoonbill.utils import common_prefix, extract_type

DOMAIN = "spoonbill"
LOCALEDIR = pkg_resources.resource_filename(DOMAIN, "locales/")
LOCALE = "en_US"

system_locale = locale.getdefaultlocale()
if system_locale and system_locale[0]:
    LOCALE = system_locale[0]


def translate(msg_id, lang=LOCALE):
    """Simple wrapper of python's gettext with ability to override desired language"""
    translator = _translation(lang)
    if translator:
        return translator.gettext(msg_id)
    return msg_id


@lru_cache(maxsize=None)
def _translation(lang):
    return gettext.translation(DOMAIN, LOCALEDIR, languages=[lang], fallback=None)


_ = translate


# slightly modified version of ocds-babel's extract_schema
# TODO: discuss upstreaming this extractor
def extract_schema_po(fileobj, keywords, comment_tags, options):
    """
    Yields json path values of a JSON Schema file.
    """
    combine = options.get("combine")
    combined_paths = []
    if combine:
        for table in combine.split(","):
            table = table.strip()
            if table in COMBINED_TABLES:
                combined_paths.extend(COMBINED_TABLES[table])
            else:
                print(f"sp_schema: {table} is not combined table!")

    def walk(prop, path="", parent_key=""):
        """Walk through schema"""
        todo = deque([(prop, "", "")])
        while todo:
            prop, path, parent_key = todo.pop()
            if prop.get("deprecated"):
                return
            properties = prop.get("properties", {})
            if properties:
                for key, item in properties.items():
                    if item.get("deprecated"):
                        continue
                    if hasattr(item, "__reference__") and item.__reference__.get("deprecated"):
                        continue
                    pointer = "/".join([path, key])
                    typeset = extract_type(item)
                    if "object" in typeset:
                        todo.append((item, pointer, key))
                    elif "array" in typeset:
                        items = item["items"]
                        items_type = extract_type(items)
                        if set(items_type) & {"array", "object"}:
                            todo.append((items, pointer, key))
                            yield pointer + "Count"
                        else:
                            yield pointer
                    else:
                        if combined_paths:
                            for p in combined_paths:
                                if common_prefix(p, pointer) == p:
                                    yield "/" + "/".join((parent_key, key))
                                    # break
                        yield pointer

    data = json.loads(fileobj.read().decode())
    schema = jsonref.JsonRef.replace_refs(data)
    for pointer in walk(schema):
        yield 1, "", pointer, ""



import collections
import logging

import xlsxwriter
from xlsxwriter.exceptions import XlsxWriterException

from spoonbill.i18n import _
from spoonbill.writers.base_writer import BaseWriter

LOGGER = logging.getLogger("spoonbill")


class XlsxWriter(BaseWriter):
    """
    Writer class with output to XLSX files.

    For each table, a corresponding sheet in an Excel workbook will be created.
    """

    name = "xlsx"

    def __init__(self, workdir, tables, options, filename="result.xlsx"):
        """
        :param workdir: Working directory
        :param tables: The table objects
        :param options: Flattening options
        """

        super().__init__(workdir, tables, options)
        self.col_index = collections.defaultdict(dict)
        self.path = workdir / filename
        self.workbook = xlsxwriter.Workbook(self.path, {"constant_memory": True})
        self.row_counters = {}

    def __enter__(self):
        """
        Write the headers to the output file.
        """

        for name, table in self.tables.items():
            table_name, headers = self.init_sheet(name, table)
            sheet = self.workbook.add_worksheet(table_name)

            for col_index, col_name in enumerate(headers):
                self.col_index[name][col_name] = col_index
                try:
                    sheet.write(0, col_index, headers[col_name])
                except XlsxWriterException as err:
                    LOGGER.error(
                        _("Failed to write header {} to xlsx sheet {} with error {}").format(col_name, name, err)
                    )
            self.row_counters[name] = 1
        return self

    def __exit__(self, *args):
        """
        Close the workbook.
        """
        LOGGER.info(_("Dumped all sheets to file to file '{}'").format(self.path))
        self.workbook.close()

    def writerow(self, table, row):
        """
        Write a row to the output file.
        """

        table_name = self.names.get(table, table)
        sheet = self.workbook.get_worksheet_by_name(table_name)
        columns = self.col_index[table]
        if not columns:
            LOGGER.error(_("Invalid table {}").format(table))
            return

        for column, value in row.items():
            if isinstance(value, bool):
                value = str(value)
            try:
                col_index = columns[column]
            except KeyError:
                LOGGER.error(
                    _("Operation produced invalid path. This a software bug, please send issue to developers")
                )
                LOGGER.error(_("Failed to write column {} to xlsx sheet {}").format(column, table))
                return
            try:
                sheet.write(self.row_counters[table], col_index, value)
            except XlsxWriterException as err:
                LOGGER.error(_("Failed to write column {} to xlsx sheet {} with error {}").format(column, table, err))

        self.row_counters[table] += 1



import json
import pathlib
from collections import OrderedDict
from decimal import Decimal

import pytest

from spoonbill.flatten import FlattenOptions
from spoonbill.spec import Column, Table
from spoonbill.stats import DataPreprocessor

from .data import TEST_COMBINED_TABLES, TEST_ROOT_TABLES

here = pathlib.Path(__file__).parent
schema_path = here / "data" / "ocds-simplified-schema.json"
releases_path = here / "data" / "ocds-sample-data.json"
analyzed_path = here / "data" / "analyzed"


@pytest.fixture
def schema():
    with open(schema_path) as fd:
        return json.load(fd, parse_float=Decimal)


@pytest.fixture
def releases():
    with open(releases_path) as fd:
        return json.load(fd, object_pairs_hook=OrderedDict)["releases"]


@pytest.fixture
def flatten_options():
    return FlattenOptions(
        **{
            "selection": {"tenders": {"split": True}, "parties": {"split": False}},
        }
    )


@pytest.fixture
def spec(schema):
    return DataPreprocessor(schema, TEST_ROOT_TABLES, combined_tables=TEST_COMBINED_TABLES)


@pytest.fixture
def spec_analyzed(schema, releases):
    dp = DataPreprocessor(schema, TEST_ROOT_TABLES, combined_tables=TEST_COMBINED_TABLES)
    for _ in dp.process_items(releases):
        pass
    return dp


@pytest.fixture
def root_table():
    return Table(
        name="tenders",
        path=["/tender"],
        is_root=True,
        columns=OrderedDict(
            [
                ("ocid", Column(title="ocid", type="string", id="ocid", hits=0)),
                ("id", Column(title="id", type="string", id="id", hits=0)),
                ("rowID", Column(title="rowID", type="string", id="rowID", hits=0)),
                (
                    "parentID",
                    Column(title="parentID", type="string", id="parentID", hits=0),
                ),
                (
                    "/tender/awardCriteriaDetails",
                    Column(
                        title="Tender Award Criteria Details",
                        type="string",
                        id="/tender/awardCriteriaDetails",
                        hits=0,
                    ),
                ),
                (
                    "/tender/items/0/id",
                    Column(
                        title="Tender Item id",
                        type="string",
                        hits=0,
                        id="/tender/items/0/id",
                    ),
                ),
                (
                    "/tender/items/0/additionalClassifications/0/id",
                    Column(
                        title="Tender Item Classification id",
                        type="string",
                        id="/tender/items/0/additionalClassifications/0/id",
                    ),
                ),
            ]
        ),
        combined_columns=OrderedDict(
            [
                ("ocid", Column(title="ocid", type="string", id="ocid", hits=0)),
                ("id", Column(title="id", type="string", id="id", hits=0)),
                ("rowID", Column(title="rowID", type="string", id="rowID", hits=0)),
                (
                    "parentID",
                    Column(title="parentID", type="string", id="parentID", hits=0),
                ),
                (
                    "/tender/awardCriteriaDetails",
                    Column(
                        title="Tender Award Criteria Details",
                        type="string",
                        id="/tender/awardCriteriaDetails",
                        hits=0,
                    ),
                ),
                (
                    "/tender/submissionMethod",
                    Column(
                        title="Tender Submission Method",
                        type="array",
                        id="/tender/submissionMethod",
                    ),
                ),
                (
                    "/tender/items/0/id",
                    Column(title="Tender Item id", type="string", id="/tender/items/0/id"),
                ),
                (
                    "/tender/items/0/additionalClassifications/0/id",
                    Column(
                        title="Tender Item Classification id",
                        type="string",
                        id="/tender/items/0/additionalClassifications/0/id",
                    ),
                ),
            ]
        ),
        arrays={"/tender/items": 0, "/tender/items/additionalClassifications": 0},
    )



from collections import defaultdict


class BaseWriter:
    """
    Base writer class
    """

    def __init__(self, workdir, tables, options):
        """
        :param workdir: Working directory
        :param tables: The table objects
        :param options: Flattening options
        """
        self.workdir = workdir
        self.tables = tables
        self.options = options
        self.names = {}
        self.headers = {}
        self.names_counter = defaultdict(int)

    def get_headers(self, table, options):
        """
        Return a table's headers, respecting the human and override options.

        :param table: A table object
        :param options: Flattening options
        :return: Mapping between the machine-readable headers and the output headers
        """
        # split = options.split and (table.roll_up or table.should_split)
        split = options.split and table.should_split
        headers = {c: c for c in table.available_rows(split=split)}
        if options.pretty_headers:
            for c in headers:
                headers[c] = table.titles.get(c, c)
        if options.headers:
            for c, h in options.headers.items():
                headers[c] = h
        return headers

    def _name_check(self, table_name):
        self.names_counter[table_name] += 1
        if self.names_counter[table_name] > 1:
            table_name += str(self.names_counter[table_name] - 1)
        return table_name

    def init_sheet(self, name, table):
        """
        Initialize a sheet, setting its headers and unique name.

        In this context, the sheet might be either a CSV file or a sheet in an Excel workbook.

        :param name: Table name
        :param table: Table object
        """
        options = self.options.selection[name]
        self.headers[name] = self.get_headers(table, options)
        self.names[name] = self._name_check(options.name or name)
        return self.names[name], self.headers[name]



"""cli.py - Command line interface related routines"""
import logging
import pathlib

import click
import click_logging

from spoonbill import FileAnalyzer, FileFlattener
from spoonbill.common import COMBINED_TABLES, ROOT_TABLES, TABLE_THRESHOLD
from spoonbill.flatten import FlattenOptions
from spoonbill.i18n import LOCALE, _
from spoonbill.utils import read_lines

LOGGER = logging.getLogger("spoonbill")
click_logging.basic_config(LOGGER)


ANALYZED_LABEL = _("  Processed {} objects")
FLATTENED_LABEL = _("  Flattened {} objects")
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


class CommaSeparated(click.ParamType):
    """Click option type to convert comma separated string into list"""

    name = "comma"

    def convert(self, value, param, ctx):  # noqa
        if not value:
            return []
        return [v for v in value.split(",")]


def read_option_file(option, option_file):
    if option_file:
        option = read_lines(option_file)
    return option


def get_selected_tables(base, selection):
    for name in selection:
        if name not in base:
            msg = _("Wrong selection, table '{}' does not exist").format(name)
            raise click.BadParameter(msg)
    return {name: tab for name, tab in base.items() if name in selection}


@click.command(context_settings=CONTEXT_SETTINGS, help=_("CLI tool to flatten OCDS files"))
@click.option(
    "--schema",
    help=_(
        "A JSON schema file URI. The URI can be a file path or an HTTP link. Spoonbill uses the schema to analyze the "
        "provided JSON file. Defaults to the OCDS 1.1.5 release schema (requires internet connection)"
    ),
    type=str,
)
@click.option(
    "--selection",
    type=CommaSeparated(),
    help=_(
        "A comma-separated list of initial tables to write. The available tables to select are: "
        "parties, planning, tenders, awards, contracts"
    ),
)
@click.option(
    "--threshold",
    help=_("The maximum number of elements in an array before it is split into a table"),
    type=int,
    default=TABLE_THRESHOLD,
    show_default=True,
)
@click.option(
    "--state-file",
    help=_("A file path URI to a previously generated state file. If not provided, a new state file is generated"),
    type=click.Path(exists=True),
)
@click.option(
    "--xlsx",
    help=_(
        "A file path to store the resulting xlsx file. Default to result.xlsx. "
        "Set to '' to disable the xlsx file generation"
    ),
    type=click.Path(),
    default="result.xlsx",
)
@click.option(
    "--csv",
    help=_("An existing directory path. If set also generates CSV files in the given directory. Disabled by default"),
    type=click.Path(),
    required=False,
)
@click.option(
    "--combine",
    help=_(
        "A comma-separated list of tables. Combines same OCDS object types from different locations "
        "(tender, awards, etc) into a single table. The available tables are: documents, milestones, and amendments"
    ),
    type=CommaSeparated(),
)
@click.option(
    "--exclude",
    help=_("A comma-separated list of tables to exclude from export. Disabled by default"),
    type=CommaSeparated(),
    default="",
)
@click.option(
    "--unnest",
    help=_(
        "A comma-separated list of column names to copy from child tables into their parent table. Disabled by default"
    ),
    type=CommaSeparated(),
    default="",
)
@click.option(
    "--unnest-file",
    help=_("A file path directory. Same as --unnest, but read column names from a file with one column per line"),
    type=click.Path(exists=True),
    required=False,
)
@click.option(
    "--only",
    help=_(
        "A comma-separated list of a subset of columns to output instead of all, in JSON path format, "
        "e.g. /parties/name. Defaults to all the available columns"
    ),
    type=CommaSeparated(),
    default="",
)
@click.option(
    "--only-file",
    help=_("A file path directory. Same as --only, but read the columns names from a file with one column per line"),
    type=click.Path(exists=True),
    required=False,
)
@click.option(
    "--repeat",
    help=_(
        "A comma-separated list of columns to repeat from a parent table into its child tables, in JSON path format,"
        "e.g. /parties/name. Disabled by default"
    ),
    type=CommaSeparated(),
    default="",
)
@click.option(
    "--repeat-file",
    help=_("A file path directory. Same as --repeat, but read the columns names from a file with one column per line"),
    type=click.Path(exists=True),
    required=False,
)
@click.option(
    "--count",
    help=_("For each array field, add a count column to its parent table. Disabled by default"),
    is_flag=True,
    default=False,
)
@click.option(
    "--human",
    help=_("Change the tables headings to human-readable format, using the schema's title properties"),
    is_flag=True,
)
@click.option(
    "--language",
    help=_("Use with --human, the language to use for the human-readable headings"),
    default=LOCALE.split("_")[0],
    show_default=True,
    type=click.Choice(["en", "es"]),
)
@click_logging.simple_verbosity_option(LOGGER)
@click.argument("filename", type=click.Path(exists=True))
def cli(
    filename,
    schema,
    selection,
    threshold,
    state_file,
    xlsx,
    csv,
    combine,
    exclude,
    unnest,
    unnest_file,
    only,
    only_file,
    repeat,
    repeat_file,
    count,
    human,
    language,
):
    """Spoonbill cli entry point"""
    if csv:
        csv = pathlib.Path(csv).resolve()
        if not csv.exists():
            raise click.BadParameter(_("Desired location {} does not exists").format(csv))
    if xlsx:
        xlsx = pathlib.Path(xlsx).resolve()
        if not xlsx.parent.exists():
            raise click.BadParameter(_("Desired location {} does not exists").format(xlsx.parent))

    path = pathlib.Path(filename)
    workdir = path.parent
    filename = path.name
    selection = selection or ROOT_TABLES.keys()
    combine = combine or COMBINED_TABLES.keys()
    root_tables = get_selected_tables(ROOT_TABLES, selection)
    combined_tables = get_selected_tables(COMBINED_TABLES, combine)

    if state_file:
        click.secho(_("Restoring from provided state file"), bold=True)
        analyzer = FileAnalyzer(workdir, state_file=state_file)
    else:
        click.secho(_("State file not supplied, going to analyze input file first"), bold=True)
        analyzer = FileAnalyzer(
            workdir,
            schema=schema,
            root_tables=root_tables,
            combined_tables=combined_tables,
            language=language,
            table_threshold=threshold,
        )
        click.echo(_("Analyze options:"))
        for name, option in ("threshold", str(threshold)), ("language", language):
            click.echo(_(" - {:30} => {}").format(name, click.style(option, fg="cyan")))
        click.echo(_("Processing file: {}").format(click.style(str(path), fg="cyan")))
        total = path.stat().st_size
        progress = 0
        # Progress bar not showing with small files
        # https://github.com/pallets/click/pull/1296/files
        with click.progressbar(width=0, show_percent=True, show_pos=True, length=total) as bar:
            for read, number in analyzer.analyze_file(filename, with_preview=True):
                bar.label = ANALYZED_LABEL.format(click.style(str(number), fg="cyan"))
                bar.update(read - progress)
                progress = read
        click.secho(
            _("Done processing. Analyzed objects: {}").format(click.style(str(number + 1), fg="red")), fg="green"
        )
        state_file = pathlib.Path(f"{filename}.state")
        state_file_path = workdir / state_file
        click.echo(_("Dumping analyzed data to '{}'").format(click.style(str(state_file_path.absolute()), fg="cyan")))
        analyzer.dump_to_file(state_file)

    click.echo(_("Flattening file: {}").format(click.style(str(path), fg="cyan")))

    if unnest and unnest_file:
        raise click.UsageError(_("Conflicting options: unnest and unnest-file"))
    if repeat and repeat_file:
        raise click.UsageError(_("Conflicting options: repeat and repeat-file"))
    if only and only_file:
        raise click.UsageError(_("Conflicting options: only and only-file"))
    if exclude:
        click.echo(_("Ignoring tables (excluded by user): {}").format(click.style(",".join(exclude), fg="red")))

    options = {"selection": {}, "count": count, "exclude": exclude}
    unnest = read_option_file(unnest, unnest_file)
    repeat = read_option_file(repeat, repeat_file)
    only = read_option_file(only, only_file)

    for name in list(selection) + list(combine):
        table = analyzer.spec[name]
        if table.total_rows == 0:
            click.echo(_("Ignoring empty table {}").format(click.style(name, fg="red")))
            continue
        options["selection"][name] = {
            "split": analyzer.spec[name].should_split,
            "pretty_headers": human,
        }
        if not analyzer.spec[name].is_combined:
            unnest_in_table = [col for col in unnest if col in table.combined_columns]
            if unnest_in_table:
                click.echo(
                    _("Unnesting columns {} for table {}").format(
                        click.style(",".join(unnest_in_table), fg="cyan"), click.style(name, fg="cyan")
                    )
                )

            only_in_table = [col for col in only if col in table]
            if only_in_table:
                click.echo(
                    _("Using only columns {} for table {}").format(
                        click.style(",".join(only_in_table), fg="cyan"), click.style(name, fg="cyan")
                    )
                )

            repeat_in_table = [col for col in repeat if col in table]
            if repeat_in_table:
                click.echo(
                    _("Repeating columns {} in all child table of {}").format(
                        click.style(",".join(repeat_in_table), fg="cyan"), click.style(name, fg="cyan")
                    )
                )
            options["selection"][name]["only"] = only_in_table
            options["selection"][name]["repeat"] = repeat_in_table
            options["selection"][name]["unnest"] = unnest_in_table

    options = FlattenOptions(**options)
    flattener = FileFlattener(
        workdir,
        options,
        analyzer,
        csv=csv,
        xlsx=xlsx,
        language=language,
    )

    click.echo(
        _("Going to export tables: {}").format(click.style(",".join(flattener.flattener.tables.keys()), fg="magenta"))
    )
    click.echo(_("Processed tables:"))
    for table_name, table in flattener.flattener.tables.items():
        msg = _(" - {:30} => {} rows") if table.is_root else _(" ---- {:27} => {} rows")
        message = msg.format(table_name, click.style(str(table.total_rows), fg="cyan"))
        click.echo(message)
    click.echo(_("Flattening input file"))
    with click.progressbar(
        flattener.flatten_file(filename),
        length=analyzer.spec.total_items + 1,
        width=0,
        show_percent=True,
        show_pos=True,
    ) as bar:
        for count in bar:
            bar.label = FLATTENED_LABEL.format(click.style(str(count + 1), fg="cyan"))

    click.secho(_("Done flattening. Flattened objects: {}").format(click.style(str(count + 1), fg="red")), fg="green")



from operator import attrgetter
from unittest.mock import call, mock_open, patch

from jmespath import search
from jsonpointer import resolve_pointer

from spoonbill.common import JOINABLE_SEPARATOR
from spoonbill.spec import Column, Table
from spoonbill.stats import DataPreprocessor
from spoonbill.utils import recalculate_headers
from tests.conftest import TEST_COMBINED_TABLES, TEST_ROOT_TABLES, schema_path
from tests.data import (
    awards_arrays,
    awards_columns,
    awards_combined_columns,
    contracts_arrays,
    contracts_columns,
    contracts_combined_columns,
    parties_arrays,
    parties_columns,
    parties_combined_columns,
    planning_arrays,
    planning_columns,
    planning_combined_columns,
    tenders_arrays,
    tenders_columns,
    tenders_combined_columns,
)

COLUMNS = {
    "tenders": tenders_columns,
    "parties": parties_columns,
    "awards": awards_columns,
    "contracts": contracts_columns,
    "planning": planning_columns,
}
COMBINED_COLUMNS = {
    "tenders": tenders_combined_columns,
    "parties": parties_combined_columns,
    "awards": awards_combined_columns,
    "contracts": contracts_combined_columns,
    "planning": planning_combined_columns,
}
ARRAYS_COLUMNS = {
    "tenders": tenders_arrays,
    "parties": parties_arrays,
    "awards": awards_arrays,
    "contracts": contracts_arrays,
    "planning": planning_arrays,
}
ADDITIONAL_COLUMNS = ["/parties/test"]


def test_parse_schema(schema, spec):
    for name in TEST_ROOT_TABLES:
        table = spec[name]
        assert isinstance(table, Table)
        assert table.total_rows == 0
        assert not table.parent
        assert table.is_root
        assert not table.is_combined
        for col_id in COLUMNS[name]:
            col = table[col_id]
            assert isinstance(col, Column)
            assert col.hits == 0
            assert col.id == col_id

        for col_id in COMBINED_COLUMNS[name]:
            col = table.combined_columns[col_id]
            assert col.hits == 0
            assert col.id == col_id

        for col_id in ARRAYS_COLUMNS[name]:
            col = table.arrays[col_id]
            assert col == 0


def test_resolve_schema_uri():
    dp = DataPreprocessor(schema_path, TEST_ROOT_TABLES, combined_tables=TEST_COMBINED_TABLES)
    assert isinstance(dp.schema, dict)


def test_get_table(spec, releases):
    table = spec.get_table("/tender")
    assert table.name == "tenders"
    table = spec.get_table("/tender/submissionMethodDetails")
    assert table.name == "tenders"
    table = spec.get_table("/tender/submissionMethod")
    assert table.name == "tenders"

    table = spec.get_table("/tender/items/id")
    assert table.name == "tenders_items"

    table = spec.get_table("/tender/items/additionalClassifications/id")
    assert table.name == "tenders_items_class"

    table = spec.get_table("/planning")
    assert table.name == "planning"
    table = spec.get_table("/parties")
    assert table.name == "parties"


# TODO: analyze combined tables
def test_analyze(spec, releases):
    [count for count in spec.process_items(releases)]
    tenders = spec.tables["tenders"]
    parties = spec.tables["parties"]
    awards = spec.tables["awards"]
    contracts = spec.tables["contracts"]

    tender_ids = search("[].tender.id", releases)
    assert tenders.total_rows == len(tender_ids)
    assert tenders["/tender/id"].hits == len(tender_ids)
    preview_ids = [i["/tender/id"] for i in tenders.preview_rows]
    assert not set(tender_ids).difference(preview_ids)

    tender_items = sorted(search("[].tender.items", releases), reverse=True, key=len)
    max_len = len(tender_items[0])
    assert tenders.arrays["/tender/items"] == max_len
    for index, item in enumerate(tender_items[0]):
        path = f"/tender/items/{index}/id"
        items = search(f"[].tender.items[{index}].id", releases)
        assert len(items) == tenders.combined_columns[path].hits
        assert len(items) == tenders.columns[path].hits
    items_ids = [i["id"] for item in tender_items for i in item]
    assert len(items_ids) == spec.tables["tenders_items"].total_rows

    for array in ("awards", "parties", "contracts"):
        ids = search(f"[].{array}[]", releases)
        table = locals().get(array)
        assert table.total_rows == len(ids)
        assert table[f"/{array}/id"].hits == len(ids)
    # check joinable calculation
    assert parties["/parties/roles"].hits == len(search("[].parties[].roles", releases))


@patch("spoonbill.LOGGER.error")
def test_mismatched_types(log, spec, releases):
    releases[0]["tender"]["id"] = ["/test/id"]
    for _ in spec.process_items(releases):
        pass
    log.assert_has_calls([call("Mismatched type on /tender/id expected ['string', 'integer']")])


@patch("spoonbill.LOGGER.error")
def test_dump_restore(log, spec, releases, tmpdir):
    for _ in spec.process_items(releases):
        pass
    spec.dump(tmpdir / "result.json")
    spec2 = DataPreprocessor.restore(tmpdir / "result.json")

    for name, table in spec.tables.items():
        assert table == spec2.tables[name]
    for key in (
        "schema",
        "root_tables",
        "combined_tables",
        "header_separator",
        "tables",
        "table_threshold",
        "total_items",
    ):
        assert key in spec2.__dict__
    with patch("builtins.open", mock_open(read_data="invalid")):
        spec2 = DataPreprocessor.restore(tmpdir / "result.json")
        log.assert_has_calls([call("Invalid pickle file. Can't restore.")])

    with patch("builtins.open", mock_open(read_data=b"invalid")):
        spec2 = DataPreprocessor.restore(tmpdir / "result.json")
        log.assert_has_calls([call("Invalid pickle file. Can't restore.")])


def test_recalculate_headers(root_table, releases):
    items = releases[0]["tender"]["items"]
    recalculate_headers(root_table, "/tender/items", "/tender", "items", items, False)
    for key in (
        "/tender/items/0/id",
        "/tender/items/0/additionalClassifications/0/id",
    ):
        assert key in root_table.combined_columns
        assert key in root_table.columns
    for key in ("/tender/items/1/id", "/tender/items/1/additionalClassifications/0/id"):
        assert key not in root_table.combined_columns
        assert key not in root_table.columns
    items = items * 2
    recalculate_headers(root_table, "/tender/items", "/tender", "items", items, False)

    for key in (
        "/tender/items/0/id",
        "/tender/items/0/additionalClassifications/0/id",
        "/tender/items/1/id",
        "/tender/items/1/additionalClassifications/0/id",
    ):
        assert key in root_table.combined_columns
        assert key in root_table.columns
    for key in ("/tender/items/2/id", "/tender/items/2/additionalClassifications/0/id"):
        assert key not in root_table.combined_columns
        assert key not in root_table.columns

    items = [
        {
            "description": "Cycle path construction work",
            "id": "45233162-2",
            "scheme": "CPV",
            "uri": "http://cpv.data.ac.uk/code-45233162.html",
        }
    ] * 2
    recalculate_headers(
        root_table,
        "/tender/items/additionalClassifications",
        "/tender/items/0",
        "additionalClassifications",
        items,
        False,
    )
    for key in (
        "/tender/items/0/id",
        "/tender/items/0/additionalClassifications/0/id",
        "/tender/items/0/additionalClassifications/1/id",
        "/tender/items/1/id",
        "/tender/items/1/additionalClassifications/0/id",
    ):
        assert key in root_table.combined_columns
        assert key in root_table.columns
    for key in (
        "/tender/items/0/additionalClassifications/2/id",
        "/tender/items/1/additionalClassifications/1/id",
        "/tender/items/2/additionalClassifications/1/id",
    ):
        assert key not in root_table.combined_columns
        assert key not in root_table.columns

    items = releases[0]["tender"]["items"] * 5
    recalculate_headers(root_table, "/tender/items", "/tender", "items", items, True)
    for key in (
        "/tender/items/0/id",
        "/tender/items/0/additionalClassifications/0/id",
        "/tender/items/1/id",
        "/tender/items/1/additionalClassifications/0/id",
        "/tender/items/2/id",
        "/tender/items/2/additionalClassifications/0/id",
        "/tender/items/3/id",
        "/tender/items/3/additionalClassifications/0/id",
        "/tender/items/4/id",
        "/tender/items/4/additionalClassifications/0/id",
    ):
        assert key in root_table.combined_columns
        assert key not in root_table.columns


def test_analyze_preview_rows(spec_analyzed, releases):
    tenders = spec_analyzed.tables["tenders"]
    tenders_items = spec_analyzed.tables["tenders_items"]
    tenders_tende = spec_analyzed.tables["tenders_tenderers"]
    tenders_items_class = spec_analyzed.tables["tenders_items_class"]

    for getter in (attrgetter("preview_rows"), attrgetter("preview_rows_combined")):
        for count, row in enumerate(getter(tenders)):
            tenderers = [r for r in getter(tenders_tende) if r["parentID"] == row["rowID"]]
            items = [r for r in getter(tenders_items) if r["parentID"] == row["rowID"]]
            items_class = []
            for it in items:
                for r in getter(tenders_items_class):
                    if r["parentID"] == it["rowID"]:
                        items_class.append(r)
            for key, item in row.items():
                if "/" in key:
                    # Check headers are present in tables
                    assert key in tenders
                    expected = resolve_pointer(releases[count], key)
                    if isinstance(expected, list):
                        # joinable
                        expected = JOINABLE_SEPARATOR.join(expected)
                    assert item == expected
                    if "tenderers" in key:
                        for index, tenderer in enumerate(reversed(tenderers)):
                            assert tenderer["parentTable"] == "tenders"
                            for k, v in tenderer.items():
                                if "/" in k:
                                    path = k.replace("/tender/tenderers", f"/tender/tenderers/{index}")
                                    value = resolve_pointer(releases[count], path)
                                    assert value == v

                    if "/tender/items/" in key:
                        for index, it in enumerate(reversed(items)):
                            assert it["parentTable"] == "tenders"
                            for k, v in it.items():
                                if "/" in k:
                                    path = k.replace("/tender/items", f"/tender/items/{index}")
                                    value = resolve_pointer(releases[count], path)
                                    assert value == v
                                    if "additionalClassifications" in k:
                                        for i, cls in enumerate(reversed(items_class)):
                                            assert cls["parentTable"] == "tenders_items"
                                            for k, v in cls.items():
                                                if "/" in k:
                                                    path = k.replace("/tender/items", f"/tender/items/{index}")
                                                    path = path.replace(
                                                        "/additionalClassifications/",
                                                        f"/additionalClassifications/{i}/",
                                                    )
                                                    value = resolve_pointer(releases[count], path)
                                                    assert v == value


def test_analyze_array_extentions_no_split(spec, releases):
    attr = {"name": "Presentacion", "id": "1"}
    items = releases[0]["tender"]["items"]
    items[0]["attributes"] = [attr]
    [_ for _ in spec.process_items(releases)]

    for cols in spec.tables["tenders"], spec.tables["tenders"].combined_columns:
        assert "/tender/items/0/attributes/0/id" in cols
        assert "/tender/items/0/attributes/0/name" in cols
        assert "/tender/items/1/attributes/0/name" not in cols
        assert "/tender/items/0/attributes/1/name" not in cols
        assert "/tender/items/1/attributes/1/name" not in cols
    for cols in spec.tables["tenders_items"], spec.tables["tenders_items"].combined_columns:
        assert "/tender/items/attributes/0/name" in cols
        assert "/tender/items/attributes/0/name" in cols
        assert "/tender/items/attributes/1/name" not in cols
        assert "/tender/items/attributes/1/name" not in cols

    for cols in spec.tables["tenders_items_attributes"], spec.tables["tenders_items_attributes"].combined_columns:
        assert "/tender/items/attributes/name" in cols
        assert "/tender/items/attributes/name" in cols


def test_analyze_array_extentions_split(spec, releases):
    attr = {"name": "Presentacion", "id": "1"}
    items = releases[0]["tender"]["items"]
    items[0]["attributes"] = [attr] * 10
    releases[0]["tender"]["items"] = items
    [_ for _ in spec.process_items(releases)]

    cols = spec.tables["tenders"]

    for key in (
        "/tender/items/0/attributes/0/id",
        "/tender/items/0/attributes/1/id",
        "/tender/items/1/attributes/0/id",
        "/tender/items/1/attributes/1/id",
        "/tender/items/0/attributes/0/name",
        "/tender/items/0/attributes/1/name",
        "/tender/items/1/attributes/0/name",
        "/tender/items/1/attributes/1/name",
    ):
        assert key not in cols

    cols = spec.tables["tenders"].combined_columns
    for key in (
        "/tender/items/0/attributes/0/id",
        "/tender/items/0/attributes/1/id",
        "/tender/items/0/attributes/0/name",
        "/tender/items/0/attributes/1/name",
    ):
        assert key in cols
    for key in (
        "/tender/items/1/attributes/0/name",
        "/tender/items/1/attributes/0/id",
        "/tender/items/1/attributes/1/name",
        "/tender/items/1/attributes/1/id",
    ):
        assert key not in cols

    cols = spec.tables["tenders_items"]
    assert "/tender/items/attributes/0/id" not in cols
    assert "/tender/items/attributes/0/name" not in cols
    assert "/tender/items/attributes/1/id" not in cols
    assert "/tender/items/attributes/1/name" not in cols

    cols = spec.tables["tenders_items"].combined_columns
    assert "/tender/items/attributes/0/id" in cols
    assert "/tender/items/attributes/0/name" in cols
    assert "/tender/items/attributes/1/id" in cols
    assert "/tender/items/attributes/1/name" in cols



import logging
from operator import attrgetter
from pathlib import Path

from ocdsextensionregistry import ProfileBuilder
from ocdskit.util import detect_format

from spoonbill.common import COMBINED_TABLES, CURRENT_SCHEMA_TAG, ROOT_TABLES, TABLE_THRESHOLD
from spoonbill.flatten import Flattener
from spoonbill.i18n import LOCALE, _
from spoonbill.stats import DataPreprocessor
from spoonbill.utils import iter_file, resolve_file_uri
from spoonbill.writers import CSVWriter, XlsxWriter

LOGGER = logging.getLogger("spoonbill")


class FileAnalyzer:
    """Main utility for analyzing files

    :param workdir: Working directory
    :param schema: Json schema file to use with data
    :param root_tables: Path configuration which should become root tables
    :param combined_tables: Path configuration for tables with multiple sources
    :param pkg_type: Field name to access records
    """

    def __init__(
        self,
        workdir,
        schema=None,
        state_file=None,
        root_tables=ROOT_TABLES,
        combined_tables=COMBINED_TABLES,
        pkg_type="releases",
        language=LOCALE,
        table_threshold=TABLE_THRESHOLD,
    ):
        self.workdir = Path(workdir)
        self.multiple_values = False
        self.schema = schema
        self.root_tables = root_tables
        self.combined_tables = combined_tables
        self.language = language
        self.table_threshold = table_threshold
        if state_file:
            self.spec = DataPreprocessor.restore(state_file)
        else:
            self.spec = None
        self.pkg_type = pkg_type

    def analyze_file(self, filename, with_preview=True):
        """Analyze provided file
        :param filename: Input filename
        :param with_preview: Generate preview during analysis
        """
        path = self.workdir / filename
        (
            input_format,
            _is_concatenated,
            _is_array,
        ) = detect_format(filename)
        LOGGER.info(_("Input file is {}").format(input_format))
        self.multiple_values = _is_concatenated
        self.parse_schema(input_format, self.schema)
        if self.spec is None:
            self.spec = DataPreprocessor(
                self.schema,
                self.root_tables,
                combined_tables=self.combined_tables,
                language=self.language,
                table_threshold=self.table_threshold,
            )
        with open(path, "rb") as fd:
            items = iter_file(fd, self.pkg_type, multiple_values=self.multiple_values)
            for count in self.spec.process_items(items, with_preview=with_preview):
                yield fd.tell(), count

    def dump_to_file(self, filename):
        """Save analyzed information to file

        :param filename: Output filename in working directory
        """
        path = self.workdir / filename
        self.spec.dump(path)

    def parse_schema(self, input_format, schema=None):
        if schema:
            schema = resolve_file_uri(schema)
        if "release" in input_format:
            pkg_type = "releases"
            getter = attrgetter("release_package_schema")
        else:
            pkg_type = "records"
            getter = attrgetter("record_package_schema")
        if not schema:
            LOGGER.info(_("No schema provided, using version {}").format(CURRENT_SCHEMA_TAG))
            profile = ProfileBuilder(CURRENT_SCHEMA_TAG, {})
            schema = getter(profile)()
        title = schema.get("title", "").lower()
        if not title:
            raise ValueError(_("Incomplete schema, please make sure your data is correct"))
        if "package" in title:
            # TODO: is is a good way to get release/record schema
            schema = schema["properties"][pkg_type]["items"]

        self.schema = schema
        self.pkg_type = pkg_type


class FileFlattener:
    """Main utility for flattening files

    :param workdir: Working directory
    :param options: Flattening configuration
    :param analyzer: Analyzed data object
    :param csv: If True generate cvs files
    :param xlsx: Generate combined xlsx table
    """

    def __init__(
        self,
        workdir,
        options,
        analyzer=None,
        tables=None,
        pkg_type="releases",
        csv=None,
        xlsx="result.xlsx",
        language=LOCALE,
        multiple_values=False,
    ):
        self.tables = tables if tables else analyzer.spec.tables
        self.flattener = Flattener(options, self.tables, language=language)
        self.workdir = Path(workdir)
        # TODO: detect package, where?
        self.writers = []
        self.csv = csv
        self.xlsx = xlsx
        self.multiple_values = multiple_values if multiple_values else analyzer.multiple_values if analyzer else False
        self.pkg_type = pkg_type if pkg_type else analyzer.pkg_type if analyzer else "releases"

    def _flatten(self, filename, writers):
        path = self.workdir / filename
        with open(path, "rb") as fd:
            items = iter_file(fd, self.pkg_type, multiple_values=self.multiple_values)
            for count, data in self.flattener.flatten(items):
                for table, rows in data.items():
                    for row in rows:
                        for wr in writers:
                            wr.writerow(table, row)
                yield count

    def flatten_file(self, filename):
        """Flatten file

        :param filename: Input filename in working directory
        """
        workdir = self.workdir
        if isinstance(self.csv, Path):
            workdir = self.csv
        if not self.xlsx and self.csv:
            with CSVWriter(workdir, self.flattener.tables, self.flattener.options) as writer:
                for count in self._flatten(filename, [writer]):
                    yield count
        if self.xlsx and not self.csv:
            with XlsxWriter(self.workdir, self.flattener.tables, self.flattener.options, filename=self.xlsx) as writer:
                for count in self._flatten(filename, [writer]):
                    yield count

        if self.xlsx and self.csv:
            with XlsxWriter(
                self.workdir, self.flattener.tables, self.flattener.options, filename=self.xlsx
            ) as xlsx, CSVWriter(workdir, self.flattener.tables, self.flattener.options) as csv:
                for count in self._flatten(filename, [xlsx, csv]):
                    yield count


__all__ = ["FileFlattener", "FileAnalyzer"]



from spoonbill.spec import add_child_table
from spoonbill.utils import combine_path, get_pointer


def test_combine_path(root_table):
    path = combine_path(root_table, "/tender/id")
    assert path == "/tender/id"
    path = combine_path(root_table, "/tender/submissionMethodDetails")
    assert path == "/tender/submissionMethodDetails"
    path = combine_path(root_table, "/tender/submissionMethod")
    assert path == "/tender/submissionMethod"
    path = combine_path(root_table, "/tender/items/id")
    assert path == "/tender/items/0/id"
    path = combine_path(root_table, "/tender/items/additionalClassifications/id")
    assert path == "/tender/items/0/additionalClassifications/0/id"
    root_table.arrays = {"/tender/items/additionalClassifications": 0}
    path = combine_path(root_table, "/tender/items/additionalClassifications/id")
    assert path == "/tender/items/additionalClassifications/0/id"


def test_inc_column(root_table):
    child = add_child_table(root_table, "/tender/items", "tender", "items")
    child.add_column("/tender/items/id", ["string"], "Tender Id", additional=True, abs_path="/tender/items/0/test")
    child.arrays["/tender/items/additionalClassifications"] = 0
    root_table.inc_column("ocid", "ocid")
    assert root_table["ocid"].hits == 1
    assert root_table.combined_columns["ocid"].hits == 1

    root_table.inc_column("/tender/awardCriteriaDetails", "/tender/awardCriteriaDetails")
    assert root_table["/tender/awardCriteriaDetails"].hits == 1
    assert root_table.combined_columns["/tender/awardCriteriaDetails"].hits == 1

    child.inc_column("/tender/items/0/id", "/tender/items/id")
    assert root_table["/tender/items/0/id"].hits == 1
    assert root_table.combined_columns["/tender/items/0/id"].hits == 1
    assert child["/tender/items/id"].hits == 1
    assert child.combined_columns["/tender/items/id"].hits == 1

    child_child = add_child_table(
        child, "/tender/items/additionalClassifications", "items", "additionalClassifications"
    )
    child_child.add_column(
        "/tender/items/additionalClassifications/id",
        ["string"],
        "Classification Id",
        additional=True,
    )
    child_child.inc_column(
        "/tender/items/0/additionalClassifications/0/id", "/tender/items/additionalClassifications/id"
    )
    assert child["/tender/items/additionalClassifications/0/id"].hits == 1
    assert child["/tender/items/additionalClassifications/0/id"].hits == 1
    assert root_table["/tender/items/0/additionalClassifications/0/id"].hits == 1
    assert root_table["/tender/items/0/additionalClassifications/0/id"].hits == 1


def test_add_column(root_table):
    root_table.add_column("/tender/id", ["string", "integer"], "Tender Id")
    assert "/tender/id" in root_table
    assert "/tender/id" in root_table.combined_columns

    root_table.add_column("/tender/itemsCount", ["string", "integer"], "Items Count")
    assert "/tender/itemsCount" in root_table
    assert "/tender/itemsCount" in root_table.combined_columns

    root_table.add_column(
        "/tender/items/additionalClassificationsCount", ["string", "integer"], "Classifications Count"
    )
    assert "/tender/items/0/additionalClassificationsCount" in root_table
    assert "/tender/items/0/additionalClassificationsCount" in root_table.combined_columns

    child = add_child_table(root_table, "/tender/items", "tender", "items")
    child.add_column(
        "/tender/items/test",
        ["string", "integer"],
        "/tender/items/test",
        additional=True,
        abs_path="/tender/items/0/test",
    )
    assert "/tender/items/test" in child
    assert "/tender/items/test" in child.combined_columns
    assert "/tender/items/0/test" in root_table
    assert "/tender/items/0/test" in root_table.combined_columns
    child.add_column("/tender/items/id", ["string", "integer"], "Items Id")

    child.arrays["/tender/items/additionalClassifications"] = 0
    child.add_column("/tender/items/additionalClassifications/id", ["string", "integer"], "Classification ID")
    assert "/tender/items/additionalClassifications/0/id" in child
    assert "/tender/items/id" in child
    assert "/tender/items/0/id" in root_table
    assert "/tender/items/0/additionalClassifications/0/id" in root_table

    child.add_column(
        "/tender/items/additionalClassificationsCount",
        ["string", "integer"],
        "Classification Count",
        propagate=False,
    )
    assert "/tender/items/additionalClassificationsCount" in child
    assert "/tender/items/additionalClassificationsCount" in child.combined_columns

    assert "/tender/items/additionalClassificationsCount" not in root_table
    assert "/tender/items/additionalClassificationsCount" not in root_table.combined_columns


def test_row_counters(root_table):
    available = ["ocid", "id", "rowID"]
    for col in available:
        root_table.inc_column(col, col)
    cols = root_table.available_rows()
    assert not set(cols).difference(available)

    cols = root_table.missing_rows()
    assert not set(cols).difference(
        [
            "parentID",
            "/tender/awardCriteriaDetails",
            "/tender/items/0/id",
            "/tender/items/0/additionalClassifications/0/id",
        ]
    )


def test_set_array(root_table):
    items = root_table.arrays["/tender/items"]
    assert items == 0

    root_table.set_array("/tender/items", [i for i in range(10)])
    items = root_table.arrays["/tender/items"]
    assert items == 10

    root_table.set_array("/tender/items", [i for i in range(5)])
    items = root_table.arrays["/tender/items"]
    assert items == 10


def test_is_array(root_table):
    assert root_table.is_array("/tender/items")
    assert root_table.is_array("/tender/items/id")
    assert root_table.is_array("/tender/items/0/id")
    assert root_table.is_array("/tender/items/additionalClassifications")
    assert not root_table.is_array("/tender/id")
    assert not root_table.is_array("/tender/title")
    assert not root_table.is_array("/tender/submissionMethod")


def test_add_child_table(root_table):
    child = add_child_table(root_table, "/tender/tenderers", "", "tenderers")
    assert child.name == "tenders_tenderers"
    assert child.name in root_table.child_tables
    assert child.total_rows == 0
    child.parent == root_table.name


def test_get_pointer(root_table):
    child = add_child_table(root_table, "/tender/items", "tender", "items")
    child_child = add_child_table(
        child, "/tender/items/additionalClassifications", "items", "additionalClassifications"
    )
    pointer = get_pointer(
        child_child,
        "/tender/items/0/additionalClassifications/0/id",
        "/tender/items/additionalClassifications/id",
        True,
    )
    assert pointer == "/tender/items/additionalClassifications/id"

    pointer = get_pointer(
        child, "/tender/items/0/additionalClassifications/0/id", "/tender/items/additionalClassifications/id", True
    )
    assert pointer == "/tender/items/additionalClassifications/0/id"

    pointer = get_pointer(
        child, "/tender/items/0/additionalClassifications/0", "/tender/items/additionalClassifications", True
    )
    assert pointer == "/tender/items/additionalClassifications/0"

    pointer = get_pointer(
        child, "/tender/items/0/additionalClassifications", "/tender/items/additionalClassifications", True
    )
    assert pointer == "/tender/items/additionalClassifications"

    pointer = get_pointer(
        root_table,
        "/tender/items/0/additionalClassifications/0/id",
        "/tender/items/additionalClassifications/id",
        True,
    )
    assert pointer == "/tender/items/0/additionalClassifications/0/id"

    pointer = get_pointer(root_table, "/tender/items/0/id", "/tender/items/id", True)
    assert pointer == "/tender/items/0/id"

    pointer = get_pointer(child, "/tender/items/0/id", "/tender/items/id", True)
    assert pointer == "/tender/items/id"
    pointer = get_pointer(root_table, "/tender/id", "/tender/id", True)
    assert pointer == "/tender/id"

    pointer = get_pointer(root_table, "/tender/items", "/tender/items", True, index="0")
    assert pointer == "/tender/items/0"

    pointer = get_pointer(root_table, "/tender", "/tender", True, index="0")
    assert pointer == "/tender"



import logging
from collections import OrderedDict
from dataclasses import dataclass, field, is_dataclass
from typing import List, Mapping, Sequence

from spoonbill.common import DEFAULT_FIELDS, DEFAULT_FIELDS_COMBINED
from spoonbill.i18n import _
from spoonbill.utils import combine_path, common_prefix, generate_table_name, get_pointer

LOGGER = logging.getLogger("spoonbill")


@dataclass
class Column:
    """
    A container for column information.

    :param title: The human-friendly title
    :param type: The expected type
    :param id: The JSON path
    :param hits: The number of times the column contains data during analysis
    """

    title: str
    type: str
    id: str
    hits: int = 0


@dataclass
class Table:
    """
    A container for table information.

    :param name: Table name
    :param path: List of paths to gather data to this table
    :param total_rows: Total available rows in this table
    :param parent: Parent table, None if this table is root table
    :param is_root: This table is root table
    :param is_combined: This table contains data collected from different paths
    :param should_split: This table should be splitted
    :param roll_up: This table should be separated from its parent
    :param columns: Columns extracted from schema for split version of this table
    :param combined_columns: Columns extracted from schema for unsplit version of this table
    :param additional_columns: Columns identified in dataset but not in schema
    :param arrays: Table array columns and maximum items (not the total count) in each array
    :param titles: All human-friendly column titles, extracted from the schema
    :param child_tables: List of possible child tables
    :param types: All paths matched to this table with corresponding object type on each path
    :param preview_rows: Generated preview for split version of this table
    :param preview_rows_combined: Generated preview for unsplit version of this table
    """

    name: str
    path: List[str]
    total_rows: int = 0
    # `parent` is a Table object, but dataclasses don't play well with recursion.
    parent: object = field(default_factory=dict)
    is_root: bool = False
    is_combined: bool = False
    should_split: bool = False
    roll_up: bool = False
    columns: Mapping[str, Column] = field(default_factory=OrderedDict)
    combined_columns: Mapping[str, Column] = field(default_factory=OrderedDict)
    additional_columns: Mapping[str, Column] = field(default_factory=OrderedDict)
    arrays: Mapping[str, int] = field(default_factory=dict)
    titles: Mapping[str, str] = field(default_factory=dict)
    child_tables: List[str] = field(default_factory=list)
    types: Mapping[str, List[str]] = field(default_factory=dict)

    preview_rows: Sequence[dict] = field(default_factory=list)
    preview_rows_combined: Sequence[dict] = field(default_factory=list)

    def __post_init__(self):
        for attr in (
            "columns",
            "combined_columns",
            "additional_columns",
        ):
            obj = getattr(self, attr, {})
            if obj:
                init = OrderedDict()
                for name, col in obj.items():
                    if not is_dataclass(col):
                        col = Column(**col)
                    init[name] = col
                setattr(self, attr, init)
            cols = DEFAULT_FIELDS_COMBINED
            if self.is_root and not self.is_combined:
                cols = DEFAULT_FIELDS
            for col in cols:
                if col not in self.columns:
                    self.columns[col] = Column(col, "string", col)
                if col not in self.combined_columns:
                    self.combined_columns[col] = Column(col, "string", col)
                self.titles[col] = _(col)

    def _counter(self, split, cond):
        cols = self.columns if split else self.combined_columns
        return [header for header, col in cols.items() if cond(col)]

    def missing_rows(self, split=True):
        """
        Return the columns that are available in the schema, but not present in the analyzed data.
        """

        return self._counter(split, lambda c: c.hits == 0)

    def available_rows(self, split=True):
        """
        Return the columns that are available in the analyzed data.
        """

        return self._counter(split, lambda c: c.hits > 0)

    def __iter__(self):
        for col in self.columns:
            yield col

    def __getitem__(self, path):
        return self.columns.get(path)

    def add_column(
        self,
        path,
        item_type,
        title,
        *,
        combined_only=False,
        propagate=True,
        additional=False,
        abs_path=None,
    ):
        """
        Add a new column to the table.

        :param path: The column's path
        :param item: Object schema description
        :param item_type: The column's expected type
        :param parent: Parent object schema description
        :param combined_only: Make this column available only in combined version of table
        :param additional: Mark this column as missing in schema
        """
        is_array = self.is_array(path)
        combined_path = combine_path(self, path)
        if not combined_only:
            self.columns[combined_path] = Column(title, item_type, combined_path)
        # new column to track hits differently
        self.combined_columns[combined_path] = Column(title, item_type, combined_path)

        if additional:
            if is_array:
                # when we analyzing file we need to keep index from data not to use 0
                # e.g. /tender/items/166/relatedLot
                combined_path = abs_path
            LOGGER.debug(_("Detected additional column: %s in %s table") % (path, self.name))
            self.additional_columns[combined_path] = Column(title, item_type, combined_path)

        for p in (path, combined_path):
            self.titles[p] = title
        if not self.is_root and propagate:
            self.parent.add_column(
                path,
                item_type,
                title,
                combined_only=combined_only,
                additional=additional,
                abs_path=abs_path,
            )

    def is_array(self, path):
        """
        Check whether the given path is in any table's arrays.
        """

        for array in sorted(self.arrays, reverse=True):
            if common_prefix(array, path) == array:
                return array
        return False

    def inc_column(self, abs_path, path):
        """
        Increment the number of non-empty cells in the column.

        :param abs_path: The column's full JSON path
        :param path: The column's JSON path without array indexes
        """
        header = get_pointer(self, abs_path, path, True)
        for headers in self.columns, self.combined_columns, self.additional_columns:
            if header in headers:
                headers[header].hits += 1

        if not self.is_root:
            self.parent.inc_column(abs_path, path)

    def add_array(self, header):
        self.arrays[header] = 0
        if not self.is_root:
            self.parent.add_array(header)

    def set_array(self, header, item):
        """
        Try to set the maximum length of an array.

        :param header: The path to the array
        :param item: Array from data
        :return: Whether the array is bigger than previously found and the length was updated
        """
        count = self.arrays.get(header, 0)
        length = len(item)
        if length > count:
            self.arrays[header] = length
            if not self.is_root:
                return self.parent.set_array(header, item)
            return True
        return False

    def inc(self):
        """
        Increment the number of rows in the table.
        """

        self.total_rows += 1
        for col_name in DEFAULT_FIELDS_COMBINED:
            self.inc_column(col_name, col_name)

    def set_preview_path(self, abs_path, path, value, max_items):
        header = get_pointer(self, abs_path, path, True)
        array = self.is_array(path)
        self.preview_rows_combined[-1][header] = value
        if header in self.combined_columns:
            if not array or (array and self.arrays[array] < max_items):
                self.preview_rows[-1][header] = value
        if not self.is_root:
            self.parent.set_preview_path(abs_path, path, value, max_items)


def add_child_table(table, pointer, parent_key, key):
    """
    Create and append a new child table to the given table.

    :param current_table: The parent table to the newly created table
    :param pointer: Path to which table should match
    :param parent_key: New table parent object filed name, used to generate table name
    :param key: New table field name object filed name, used to generate table name
    :return: Child table
    """
    table_name = generate_table_name(table.name, parent_key, key)
    child_table = Table(table_name, [pointer], parent=table)
    table.child_tables.append(table_name)
    table.add_array(pointer)
    return child_table



import logging
import pickle
from collections import defaultdict, deque
from functools import lru_cache
from pathlib import Path
from typing import List, Mapping

import jsonref
from flatten_dict import flatten

from spoonbill.common import ARRAY, JOINABLE, JOINABLE_SEPARATOR, PREVIEW_ROWS, TABLE_THRESHOLD
from spoonbill.i18n import LOCALE, _
from spoonbill.rowdata import Rows
from spoonbill.spec import Table, add_child_table
from spoonbill.utils import (
    PYTHON_TO_JSON_TYPE,
    RepeatFilter,
    extract_type,
    generate_table_name,
    get_matching_tables,
    get_root,
    recalculate_headers,
    resolve_file_uri,
    validate_type,
)

LOGGER = logging.getLogger("spoonbill")
LOGGER.addFilter(RepeatFilter())


class DataPreprocessor:
    """
    Data analyzer

    Processes the given schema and, based on this, extracts information from the iterable dataset.

    :param schema: The dataset's schema
    :param root_tables: The paths which should become root tables
    :param combined_tables: The paths which should become tables that combine data from different locations
    :param tables: Use these tables objects instead of parsing the schema
    :param table_threshold: The maximum array length, before it is recommended to split out a child table
    :param total_items: The total objects processed
    """

    def __init__(
        self,
        schema: Mapping,
        root_tables: Mapping[str, List],
        combined_tables: Mapping[str, List] = None,
        tables: Mapping[str, Table] = None,
        table_threshold=TABLE_THRESHOLD,
        total_items=0,
        header_separator="/",
        language=LOCALE,
    ):
        self.schema = schema
        self.root_tables = root_tables
        self.combined_tables = combined_tables or {}
        self.tables = tables or {}
        self.table_threshold = table_threshold

        self.header_separator = header_separator
        self.total_items = total_items
        self.current_table = None

        self.language = language
        self.names_counter = defaultdict(int)
        if not self.tables:
            self.parse_schema()

    def __getitem__(self, table):
        return self.tables[table]

    def name_check(self, parent_key, key):
        table_name = generate_table_name(self.current_table.name, parent_key, key)
        self.names_counter[table_name] += 1
        if self.names_counter[table_name] > 1:
            key = key[:4] + str(self.names_counter[table_name] - 1)
        return key

    def init_tables(self, tables, is_combined=False):
        """
        Initialize the root tables with default fields.
        """
        for name, path in tables.items():
            table = Table(name, path, is_root=True, is_combined=is_combined, parent="")
            self.tables[name] = table

    def parse_schema(self):
        """
        Extract information from the schema.
        """
        if isinstance(self.schema, (str, Path)):
            self.schema = resolve_file_uri(self.schema)
        self.schema = jsonref.JsonRef.replace_refs(self.schema)
        self.init_tables(self.root_tables)
        if self.combined_tables:
            self.init_tables(self.combined_tables, is_combined=True)
        separator = self.header_separator
        to_analyze = deque([("", "", {}, self.schema)])

        # TODO: check if recursion is better for field ordering
        while to_analyze:
            path, parent_key, parent, prop = to_analyze.pop()
            if prop.get("deprecated"):
                continue
            # TODO: handle oneOf anyOf allOf
            properties = prop.get("properties", {})
            if properties:
                for key, item in properties.items():
                    if item.get("deprecated"):
                        continue
                    if hasattr(item, "__reference__") and item.__reference__.get("deprecated"):
                        continue
                    typeset = extract_type(item)
                    pointer = separator.join([path, key])
                    self.current_table = self.get_table(pointer)
                    if not self.current_table:
                        continue

                    self.current_table.types[pointer] = typeset
                    if "object" in typeset:
                        to_analyze.append((pointer, key, properties, item))
                    elif "array" in typeset:
                        items = item["items"]
                        items_type = extract_type(items)
                        if set(items_type) & {"array", "object"}:
                            if pointer not in self.current_table.path:
                                # found child array, need to create child table
                                key = self.name_check(parent_key, key)
                                self._add_table(add_child_table(self.current_table, pointer, parent_key, key), pointer)
                            to_analyze.append((pointer, key, properties, items))
                        else:
                            # This means we in array of strings, so this becomes a single joinable column
                            typeset = ARRAY.format(items_type)
                            self.current_table.types[pointer] = JOINABLE
                            self.current_table.add_column(pointer, typeset, _(pointer, self.language))
                    else:
                        if self.current_table.is_combined:
                            pointer = separator + separator.join((parent_key, key))
                        self.current_table.add_column(pointer, typeset, _(pointer, self.language))
            else:
                # TODO: not sure what to do here
                continue

    def _add_table(self, table, pointer):
        self.tables[table.name] = table
        self.current_table = table
        self.get_table.cache_clear()

    def _add_additional_table(self, pointer, abs_pointer, parent_key, key, item, separator="/"):
        LOGGER.debug(_("Detected additional table: %s") % pointer)
        self.current_table.types[pointer] = ["array"]
        self._add_table(add_child_table(self.current_table, pointer, parent_key, key), pointer)
        # add columns beforehand because it might be required
        # to recalculate  and reorder headers when enlarging array
        # there must be a better way but it should work for now
        for extended_item in item:
            for path_, it in flatten(extended_item, reducer="path").items():
                p = separator.join((pointer, path_))
                a_p = separator.join((abs_pointer, path_))
                if p not in self.current_table:
                    self.current_table.add_column(
                        p,
                        PYTHON_TO_JSON_TYPE.get(type(it).__name__, "N/A"),
                        _(p, self.language),
                        abs_path=a_p,
                    )

    @lru_cache(maxsize=None)
    def get_table(self, path):
        """
        Get the table that best matches the given path.

        :param path: A path
        :return: A table
        """
        candidates = get_matching_tables(self.tables, path)
        if not candidates:
            return
        return candidates[0]

    def add_preview_row(self, rows, item_id, parent_key):
        """
        Append a mostly-empty row to the previews.

        This is important to do, because other code uses an index of -1 to access and update the current row.

        :param ocid: The row's ocid
        :param item_id: The current object's id
        :param parent_id: The parent object's id
        :param parent_table: The parent table's name
        :param buyer: The Buyer object if available
        """
        table = self.current_table
        for p_rows in table.preview_rows, table.preview_rows_combined:
            p_rows.append(rows.new_row(table, item_id).as_dict())

    def process_items(self, releases, with_preview=True):
        """
        Analyze releases.

        Iterates over every release to calculate metrics and optionally generates previews for combined and split
        versions of each table.

        :param releases: The releases to analyze
        :param with_preview: Whether to generate previews for each table
        """
        separator = self.header_separator
        for count, release in enumerate(releases):
            to_analyze = deque([("", "", "", {}, release)])
            rows = Rows(ocid=release["ocid"], buyer=release.get("buyer", {}), data=defaultdict(list))

            while to_analyze:
                abs_path, path, parent_key, parent, record = to_analyze.pop()
                for key, item in record.items():
                    pointer = separator.join([path, key])
                    self.current_table = self.get_table(pointer)
                    if not self.current_table:
                        continue
                    item_type = self.current_table.types.get(pointer)
                    if pointer in self.current_table.path and pointer != "/buyer":
                        # strict match like /parties, /tender
                        c = item if isinstance(item, list) else [item]
                        for _nop in c:
                            self.current_table.inc()
                            if with_preview and count < PREVIEW_ROWS:
                                self.add_preview_row(rows, record.get("id", ""), parent_key)

                    # TODO: this validation should probably be smarter with arrays
                    if item_type and item_type != JOINABLE and not validate_type(item_type, item):
                        LOGGER.error("Mismatched type on %s expected %s" % (pointer, item_type))
                        continue

                    if isinstance(item, dict):
                        to_analyze.append(
                            (
                                separator.join([abs_path, key]),
                                pointer,
                                key,
                                record,
                                item,
                            )
                        )
                    elif item and isinstance(item, list):
                        abs_pointer = separator.join([abs_path, key])
                        if not isinstance(item[0], dict) and not item_type:
                            LOGGER.debug(
                                _("Detected additional column: %s in %s table")
                                % (abs_pointer, self.current_table.name)
                            )
                            item_type = JOINABLE
                            self.current_table.add_column(
                                pointer,
                                JOINABLE,
                                _(pointer, self.language),
                                additional=True,
                                abs_path=abs_pointer,
                            )
                        if item_type == JOINABLE:
                            self.current_table.inc_column(abs_pointer, pointer)
                            if with_preview and count < PREVIEW_ROWS:
                                value = JOINABLE_SEPARATOR.join(item)
                                self.current_table.set_preview_path(abs_pointer, pointer, value, self.table_threshold)
                        elif self.current_table.is_root or self.current_table.is_combined:
                            for value in item:
                                to_analyze.append(
                                    (
                                        abs_pointer,
                                        pointer,
                                        key,
                                        record,
                                        value,
                                    )
                                )
                        else:
                            parent_table = self.current_table.parent
                            if pointer not in parent_table.arrays:
                                LOGGER.debug(_("Detected additional table: %s") % pointer)
                                self.current_table.types[pointer] = ["array"]
                                parent_table = self.current_table
                                # TODO: do we need to mark this table as additional
                                self._add_additional_table(pointer, abs_pointer, parent_key, key, item)
                                self.add_preview_row(rows, record.get("id", ""), parent_key)
                            if parent_table.set_array(pointer, item):
                                should_split = len(item) >= self.table_threshold
                                if should_split:
                                    parent_table.should_split = True
                                    self.current_table.roll_up = True
                                recalculate_headers(
                                    parent_table, pointer, abs_path, key, item, should_split, separator
                                )

                            for i, value in enumerate(item):
                                if isinstance(value, dict):
                                    abs_pointer = separator.join([abs_path, key, str(i)])
                                    to_analyze.append(
                                        (
                                            abs_pointer,
                                            pointer,
                                            parent_key,
                                            record,
                                            value,
                                        )
                                    )
                    else:
                        root = get_root(self.current_table)
                        abs_pointer = separator.join((abs_path, key))
                        if self.current_table.is_combined:
                            pointer = separator + separator.join((parent_key, key))
                            abs_pointer = pointer
                        if abs_pointer not in root.combined_columns:
                            self.current_table.add_column(
                                pointer,
                                PYTHON_TO_JSON_TYPE.get(type(item).__name__, "N/A"),
                                _(pointer, self.language),
                                additional=True,
                                abs_path=abs_pointer,
                            )
                        self.current_table.inc_column(abs_pointer, pointer)
                        if item and with_preview and count < PREVIEW_ROWS:
                            if not pointer.startswith("/buyer"):
                                self.current_table.set_preview_path(abs_pointer, pointer, item, self.table_threshold)
            yield count
        self.total_items = count

    def dump(self, path):
        """
        Dump the data processor's state to a file.

        :param path: Full path to file
        """
        try:
            with open(path, "wb") as fd:
                pickle.dump(self, fd)
        except (OSError, IOError) as e:
            LOGGER.error(_("Failed to dump DataPreprocessor to file. Error: {}").format(e))

    @classmethod
    def restore(_cls, path):
        """
        Restore a data preprocessor's state from a file.

        :param path: Full path to file
        """
        try:
            with open(path, "rb") as fd:
                return pickle.load(fd)
        except (TypeError, pickle.UnpicklingError):
            LOGGER.error(_("Invalid pickle file. Can't restore."))



import codecs
import functools
import json
import logging
from collections import OrderedDict
from dataclasses import replace
from itertools import chain
from numbers import Number
from pathlib import Path

import ijson
import requests

PYTHON_TO_JSON_TYPE = {
    "list": "array",
    "dict": "object",
    "string": "string",
    "int": "integer",
    "float": "number",
}
LOGGER = logging.getLogger("spoonbill")

ABBREVIATION_KEY = {
    "additionalIdentifiers": "ids",
    "additionalClassifications": "class",
    "documents": "docs",
}

ABBREVIATION_TABLE_NAME = {
    "contracts_implementation": "implementation",
    "contracts_implementation_transactions": "transactions",
}


@functools.lru_cache(maxsize=None)
def common_prefix(path, subpath, separator="/"):
    """Given two paths, returns the longest common sub-path.

    >>> common_prefix('/contracts', '/contracts/items')
    '/contracts'
    >>> common_prefix('/tender/submissionMethod', '/tender/submissionMethodDetails')
    '/tender'
    >>> common_prefix('/tender/items/id', '/tender/items/description')
    '/tender/items'
    >>> common_prefix('/tender/items/0/additionalClassifications/0/id', '/tender/items/0')
    '/tender/items/0'
    """
    paths = [path.split(separator), subpath.split(separator)]
    if len(paths[0]) <= len(paths[1]):
        s1, s2 = paths
    else:
        s2, s1 = paths
    for i, path in enumerate(s1):
        if path != s2[i]:
            common = s1[:i]
            break
    else:
        common = s1
    return separator.join(common)


def iter_file(fd, root, multiple_values=False):
    """Iterate over `root` array in file provided by `filename` using ijson

    :param bytes fd: File descriptor
    :param str root: Array field name inside file
    :param bool multiple_values: Determine line-delimited JSON
    :return: Iterator of bytes read and item as a tuple

    >>> [r for r in iter_file(open('tests/data/ocds-sample-data.json', 'rb'), 'records')]
    []
    >>> len([r for r in iter_file(open('tests/data/ocds-sample-data.json', 'rb'), 'releases')])
    6
    """
    reader = ijson.items(
        fd, prefix=("" if multiple_values else f"{root}.item"), multiple_values=multiple_values, map_type=OrderedDict
    )
    for item in reader:
        yield item


def extract_type(item):
    """Extract item possible types from jsonschema definition.
    >>> extract_type({'type': 'string'})
    ['string']
    >>> extract_type(None)
    []
    >>> extract_type({})
    []
    >>> extract_type({'type': ['string', 'null']})
    ['string', 'null']
    """
    if not item or "type" not in item:
        return []
    type_ = item["type"]
    if not isinstance(type_, list):
        type_ = [type_]
    return type_


def validate_type(type_, item):
    """Validate if python object corresponds to provided type
    >>> validate_type(['string'], 'test_string')
    True
    >>> validate_type(['number'], 11.1)
    True
    >>> validate_type(['number'], 11)
    True
    >>> validate_type(['array'], [])
    True
    >>> validate_type(['array'], {})
    False
    >>> validate_type(['object'], [])
    False
    >>> validate_type(['object'], {})
    True
    """
    if isinstance(item, Number):
        name = "number"
    else:
        name = type(item).__name__
    expected = PYTHON_TO_JSON_TYPE.get(name)
    if expected:
        return expected in type_
    return True


def get_root(table):
    """Extract top level toot table of `table`"""
    while table.parent:
        table = table.parent
    return table


def combine_path(root, path, index="0", separator="/"):
    """Generates index based header for combined column"""
    combined_path = path
    for array in sorted(root.arrays, reverse=True):
        if common_prefix(path, array) == array:
            chunk = separator.join((array, index))
            combined_path = combined_path.replace(array, chunk)
    return combined_path


def get_matching_tables(tables, path):
    """Get list of matching tables for provided path

    Return list is sorted by longest matching path part

    :param tables: List of `Table' objects
    :param path: Path like string
    :return: List of matched by path tables
    """
    candidates = []
    for table in tables.values():
        for candidate in table.path:
            if common_prefix(candidate, path) == candidate:
                candidates.append(table)

    return sorted(candidates, key=lambda c: max((len(p) for p in c.path)), reverse=True)


def generate_table_name(parent_table, parent_key, key):
    """Generates name for non root table, to be used as sheet name

    :param str parent_table: Parent table name
    :param str parent_key: Parent object field name
    :param str key: Current object field name
    :return: Generated table name
    :rtype: str

    >>> generate_table_name('tenders', 'tender', 'items')
    'tenders_items'
    >>> generate_table_name('tenders', 'items', 'additionalClassifications')
    'tenders_items_class'
    >>> generate_table_name('parties', 'parties', 'roles')
    'parties_roles'
    """

    if key in ABBREVIATION_KEY:
        key = ABBREVIATION_KEY[key]

    if parent_key in parent_table:
        table_name = f"{parent_table}_{key}"
    else:
        table_name = f"{parent_table}_{parent_key}_{key}"

    if table_name in ABBREVIATION_TABLE_NAME:
        table_name = ABBREVIATION_TABLE_NAME[table_name]

    if len(table_name) >= 31:
        if parent_key in parent_table:
            table_name = f"{parent_table}_{key[:5]}"
        else:
            table_name = f"{parent_table}_{parent_key[:5]}_{key[:5]}"

    return table_name


def recalculate_headers(table, path, abs_path, key, item, should_split, separator="/"):
    """Rebuild table headers when array is expanded with attempt to preserve order

    Also deletes combined columns from tables columns if array becomes bigger than threshold

    :param table: Table for which headers should be rebuild
    :param abs_path: Full jsonpath to array
    :param key: Array field name
    :param item: Array items
    :param should_split: True if array should be separated into child table
    :param separator: header path separator
    """

    def insert_after_key(columns, insert, last_key):
        data = OrderedDict()
        for key, val in columns.items():
            data[key] = val
            if key == last_key:
                for k, v in insert.items():
                    data[k] = v
                    table.titles[k] = v.title
        return data

    base_prefix = separator.join((abs_path, key))
    zero_prefix = get_pointer(table, separator.join((base_prefix, "0")), path, True)

    zero_cols = {
        col_p: col
        for col_p, col in table.combined_columns.items()
        if col_p.startswith(separator) and common_prefix(col_p, zero_prefix) == zero_prefix
    }
    new_cols = {}
    for col_i, _ in enumerate(item[1:], 1):
        col_prefix = get_pointer(table, separator.join((base_prefix, str(col_i))), path, True)

        for col_p, col in zero_cols.items():
            col_id = col.id.replace(zero_prefix, col_prefix)
            new_cols[col_id] = replace(col, id=col_id, hits=0)

    if new_cols:
        last_key = list(zero_cols.keys())[-1]
        table.combined_columns = insert_after_key(table.combined_columns, new_cols, last_key)
        if should_split:
            for col_path in chain(zero_cols, new_cols):
                table.columns.pop(col_path, "")
        else:
            table.columns = insert_after_key(table.columns, new_cols, last_key)
        if not table.is_root:
            recalculate_headers(table.parent, path, abs_path, key, item, should_split, separator)


def resolve_file_uri(file_path):
    """Read json file from provided uri

    :param file_path: URI to file, could be url or path
    :return: Read file as dictionary
    """
    if isinstance(file_path, (str, Path)):
        with codecs.open(file_path, encoding="utf-8") as fd:
            return json.load(fd, object_pairs_hook=OrderedDict)
    if file_path.startswith("http://") or file_path.startswith("https://"):
        return requests.get(file_path).json()


def read_lines(path):
    """Read file as lines"""
    with open(path) as fd:
        return [line.strip() for line in fd.readlines()]


def get_pointer(table, abs_path, path, split, *, separator="/", index=None):
    """Combine path and abs_path in order to fit table columns

    For example /tender/items/0/id should be /tender/items/0/id for tenders table
    but /tender/items/id for tenders_items table
    """
    array = table.is_array(path)
    if index and (array or table.is_combined):
        return separator.join((abs_path, index))
    if table.is_root:
        return abs_path

    if array:
        paths = abs_path.split(separator)
        prefix = ""

        for index, pth in enumerate(paths, 1):
            if pth.isdigit():
                continue
            if not pth:
                continue
            prefix = separator.join((prefix, pth))
            if prefix == array:
                break
        pointer = separator.join(paths[index:])
        if pointer:
            return separator.join((prefix, pointer))
        return prefix
    return path


class RepeatFilter(logging.Filter):
    """
    Logger filter to avoid repeating of same messages during file processing
    """

    def filter(self, record):
        current_log = (record.module, record.levelno, record.msg)
        if current_log != getattr(self, "last_log", None):
            self.last_log = current_log
            return True
        return False


def make_count_column(array):
    """Make column name for arrays elements count

    >>> make_count_column('/tender/items')
    '/tender/itemsCount'
    >>> make_count_column('/tender/items/additionalClassifications')
    '/tender/items/additionalClassificationsCount'
    >>> make_count_column('/tender/items/')
    '/tender/itemsCount'
    """

    return array.rstrip("/") + "Count"



import csv
from pathlib import Path
from unittest.mock import call, patch

import openpyxl

from spoonbill import FileAnalyzer, FileFlattener
from spoonbill.flatten import Flattener, FlattenOptions
from spoonbill.writers.csv import CSVWriter
from spoonbill.writers.xlsx import XlsxWriter

from .conftest import releases_path
from .utils import get_writers, prepare_tables, read_csv_headers, read_xlsx_headers

ID_FIELDS = {"tenders": "/tender/id", "parties": "/parties/id"}


def test_writer_init(spec, tmpdir, flatten_options):
    tables = prepare_tables(spec, flatten_options, ID_FIELDS)
    workdir = Path(tmpdir)
    get_writers(workdir, tables, flatten_options)
    path = workdir / "result.xlsx"

    assert path.is_file()
    for name in flatten_options.selection:
        path = workdir / f"{name}.csv"
        assert path.is_file()


def test_headers_filtering(spec, tmpdir, flatten_options):
    tables = prepare_tables(spec, flatten_options, ID_FIELDS)
    workdir = Path(tmpdir)
    get_writers(workdir, tables, flatten_options)
    for name in flatten_options.selection:
        path = workdir / f"{name}.csv"
        xlsx_headers = read_xlsx_headers(workdir / "result.xlsx", name)
        csv_headers = read_csv_headers(path)
        assert len(xlsx_headers) == 1
        assert len(csv_headers) == 1
        assert xlsx_headers[0] == ID_FIELDS[name]
        assert csv_headers[0] == ID_FIELDS[name]


def test_writers_pretty_headers(spec, tmpdir, releases):
    # increase items count for force split
    releases[0]["tender"]["items"] = releases[0]["tender"]["items"] * 6
    for _ in spec.process_items(releases):
        pass
    options = FlattenOptions(
        **{
            "selection": {
                "tenders": {"split": True, "pretty_headers": True},
                "parties": {"split": False, "pretty_headers": True},
                "tenders_items": {
                    "split": True,
                    "headers": {"/tender/items/id": "item id"},
                    "pretty_headers": True,
                },
            }
        }
    )
    tables = {
        "tenders": spec.tables["tenders"],
        "parties": spec.tables["parties"],
        "tenders_items": spec.tables["tenders_items"],
    }

    workdir = Path(tmpdir)
    get_writers(workdir, tables, options)
    xlsx = workdir / "result.xlsx"

    for name, opts in options.selection.items():
        path = workdir / f"{name}.csv"
        xlsx_headers = read_xlsx_headers(xlsx, name)
        csv_headers = read_csv_headers(path)
        table = tables[name]
        for col in tables[name].available_rows(opts.split):
            title = table.titles.get(col)
            if col == "/tender/items/id":
                title = "item id"
            assert title in xlsx_headers
            assert title in csv_headers

    options = FlattenOptions(
        **{
            "selection": {
                "tenders": {
                    "split": True,
                    "headers": {"/tender/id": "TENDER"},
                    "pretty_headers": True,
                },
                "tenders_items": {
                    "split": True,
                    "headers": {"/tender/items/id": "item id"},
                    "pretty_headers": True,
                },
                "parties": {
                    "split": False,
                    "headers": {"/parties/id": "PARTY"},
                    "pretty_headers": True,
                },
            }
        }
    )

    workdir = Path(tmpdir)
    get_writers(workdir, tables, options)
    xlsx = workdir / "result.xlsx"

    sheet = "tenders"
    path = workdir / f"{sheet}.csv"
    xlsx_headers = read_xlsx_headers(xlsx, sheet)
    csv_headers = read_csv_headers(path)
    assert "TENDER" in xlsx_headers
    assert "TENDER" in csv_headers

    sheet = "tenders_items"
    path = workdir / f"{sheet}.csv"
    xlsx_headers = read_xlsx_headers(xlsx, sheet)
    csv_headers = read_csv_headers(path)
    assert "item id" in xlsx_headers
    assert "item id" in csv_headers

    xlsx = workdir / "result.xlsx"
    sheet = "parties"
    path = workdir / f"{sheet}.csv"

    xlsx_headers = read_xlsx_headers(xlsx, sheet)
    csv_headers = read_csv_headers(path)

    assert "PARTY" in xlsx_headers
    assert "PARTY" in csv_headers


def test_writers_flatten_count(spec, tmpdir, releases):
    releases[0]["tender"]["items"] = releases[0]["tender"]["items"] * 6
    for _ in spec.process_items(releases):
        pass
    options = FlattenOptions(
        **{
            "selection": {
                "tenders": {"split": True, "pretty_headers": True},
                "parties": {"split": True, "pretty_headers": True},
                "tenders_items": {
                    "split": False,
                    "pretty_headers": True,
                },
            },
            "count": True,
        }
    )

    workdir = Path(tmpdir)
    analyzer = FileAnalyzer(workdir)
    flattener = FileFlattener(workdir=workdir, options=options, tables=spec.tables, csv=True, analyzer=analyzer)
    xlsx = workdir / "result.xlsx"
    for _ in flattener.flatten_file(releases_path):
        pass
    sheet = "tenders"
    path = workdir / f"{sheet}.csv"
    for headers in read_xlsx_headers(xlsx, sheet), read_csv_headers(path):
        assert "Items Count" in headers
        assert "Tenderers Count" in headers

    sheet = "parties"
    path = workdir / f"{sheet}.csv"
    for headers in read_xlsx_headers(xlsx, sheet), read_csv_headers(path):
        assert "Additional Identifiers Count" in headers


def test_writers_table_name_override(spec, tmpdir):
    options = FlattenOptions(
        **{
            "selection": {
                "parties": {"split": False, "pretty_headers": True, "name": "testname"},
                "tenders": {"split": True, "pretty_headers": True, "name": "my_tenders"},
                "tenders_items": {"split": False, "pretty_headers": True, "name": "pretty_items"},
            }
        }
    )
    tables = prepare_tables(spec, options)
    for name, table in tables.items():
        for col in table:
            table.inc_column(col, col)

    workdir = Path(tmpdir)
    get_writers(workdir, tables, options)
    xlsx = workdir / "result.xlsx"
    for name in ("testname", "my_tenders", "pretty_items"):
        path = workdir / f"{name}.csv"
        assert path.is_file()
        assert read_xlsx_headers(xlsx, name)
        assert read_csv_headers(path)


def test_abbreviations(spec, tmpdir):
    options = FlattenOptions(
        **{
            "selection": {
                "tenders_items_class": {"split": False},
                "parties_ids": {"split": False},
                "transactions": {"split": False},
            }
        }
    )
    new_names = ["tenders_items_class", "parties_ids", "transactions"]
    tables = prepare_tables(spec, options)
    for name, table in tables.items():
        for col in table:
            table.inc_column(col, col)

    workdir = Path(tmpdir)
    get_writers(workdir, tables, options)
    xlsx = workdir / "result.xlsx"
    for name in new_names:
        path = workdir / f"{name}.csv"
        assert path.is_file()
        assert read_xlsx_headers(xlsx, name)
        assert read_csv_headers(path)


def test_name_duplicate(spec, tmpdir):
    duplicate_name = "test"
    options = FlattenOptions(
        **{
            "selection": {
                "parties": {"split": False, "pretty_headers": True, "name": duplicate_name},
                "tenders": {"split": True, "pretty_headers": True, "name": duplicate_name},
                "tenders_items": {"split": False, "pretty_headers": True, "name": duplicate_name},
            }
        }
    )
    tables = prepare_tables(spec, options)
    for name, table in tables.items():
        for col in table:
            table.inc_column(col, col)
    workdir = Path(tmpdir)
    get_writers(workdir, tables, options)
    xlsx = workdir / "result.xlsx"
    for name in ("test", "test1", "test2"):
        path = workdir / f"{name}.csv"
        assert path.is_file()
        assert read_xlsx_headers(xlsx, name)
        assert read_csv_headers(path)


@patch("spoonbill.LOGGER.error")
def test_writers_invalid_table(log, spec, tmpdir):
    options = FlattenOptions(
        **{
            "selection": {
                "parties": {"split": False, "pretty_headers": True, "name": "testname"},
            }
        }
    )
    tables = prepare_tables(spec, options)
    for name, table in tables.items():
        for col in table:
            table.inc_column(col, col)

    workdir = Path(tmpdir)
    writers = get_writers(workdir, tables, options)
    for writer in writers:
        writer.writerow("test", {})
    log.assert_has_calls([call("Invalid table test"), call("Invalid table test")])


@patch("spoonbill.LOGGER.error")
def test_writers_invalid_row(log, spec, tmpdir):
    options = FlattenOptions(
        **{
            "selection": {
                "parties": {"split": False, "pretty_headers": True, "name": "testname"},
            }
        }
    )
    tables = prepare_tables(spec, options)
    for name, table in tables.items():
        for col in table:
            table.inc_column(col, col)

    workdir = Path(tmpdir)
    writers = get_writers(workdir, tables, options)
    for writer in writers:
        writer.writerow("parties", {"/test/test": "test"})
    log.assert_has_calls(
        [
            call("Operation produced invalid path. This a software bug, please send issue to developers"),
            call("Failed to write row None with error dict contains fields not in fieldnames: '/test/test'"),
            call("Operation produced invalid path. This a software bug, please send issue to developers"),
            call("Failed to write column /test/test to xlsx sheet parties"),
        ]
    )


@patch("spoonbill.LOGGER.error")
@patch("spoonbill.writers.csv.open", **{"side_effect": OSError("test")})
def test_writers_open_fail(open_, log, spec, tmpdir):
    options = FlattenOptions(
        **{
            "selection": {
                "parties": {"split": False, "pretty_headers": True, "name": "testname"},
            }
        }
    )
    workdir = Path(tmpdir)
    tables = prepare_tables(spec, options)
    get_writers(workdir, tables, options)
    log.assert_has_calls([call("Failed to open file {} with error {}".format(str(tmpdir / "testname.csv"), "test"))])


def test_csv_writer(spec_analyzed, releases, flatten_options, tmpdir):
    flattener = Flattener(flatten_options, spec_analyzed.tables)
    tables = prepare_tables(spec_analyzed, flatten_options)
    workdir = Path(tmpdir)
    with CSVWriter(workdir, tables, flatten_options) as writer:
        # Writing CSV files
        for _count, flat in flattener.flatten(releases):
            for name, rows in flat.items():
                for row in rows:
                    writer.writerow(name, row)

    # Reading CSV files
    counter = {}
    for _count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            if name not in counter:
                counter[name] = 0
            for row in rows:
                str_row = {k: str(v) for (k, v) in row.items()}
                file = name + ".csv"
                path = workdir / file
                with open(path, newline="") as csv_file:
                    csv_reader = csv.DictReader(csv_file)
                    for num, line in enumerate(csv_reader):
                        if num == counter[name]:
                            clean_line = {k: v for (k, v) in line.items() if v != ""}
                            assert dict(clean_line) == str_row
                counter[name] += 1


def test_xlsx_writer(spec_analyzed, releases, flatten_options, tmpdir):
    flattener = Flattener(flatten_options, spec_analyzed.tables)
    tables = prepare_tables(spec_analyzed, flatten_options)
    workdir = Path(tmpdir)
    with XlsxWriter(workdir, tables, flatten_options) as writer:
        # Writing XLSX file
        for _count, flat in flattener.flatten(releases):
            for name, rows in flat.items():
                for row in rows:
                    writer.writerow(name, row)

    # Reading XLSX files
    counter = {}
    path = workdir / "result.xlsx"
    for _count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            if name not in counter:
                counter[name] = 2
            xlsx_reader = openpyxl.load_workbook(path)
            sheet = xlsx_reader[name]
            header_values = [cell.value for cell in sheet[1]]
            header_columns = [cell.column_letter for cell in sheet[1]]
            headers = dict(zip(header_columns, header_values))
            for row in rows:
                line_values = [cell.value for cell in sheet[counter[name]]]
                line_columns = [headers[cell.column_letter] for cell in sheet[counter[name]]]
                line = dict(zip(line_columns, line_values))
                # Cleaning empty cells
                line = {k: v for (k, v) in line.items() if v}

                if "/tender/hasEnquiries" in row:
                    str_row = {k: v for (k, v) in row.items()}
                    str_row["/tender/hasEnquiries"] = str(row["/tender/hasEnquiries"])
                    assert line == str_row
                else:
                    assert line == row.as_dict()
                counter[name] += 1


def test_less_five_arrays_csv(spec_analyzed, releases, flatten_options, tmpdir):
    test_arrays = ["tenders_items", "tenders_items_addit", "tenders_tende"]
    flattener = Flattener(flatten_options, spec_analyzed.tables)
    tables = prepare_tables(spec_analyzed, flatten_options)
    workdir = Path(tmpdir)
    with CSVWriter(workdir, tables, flatten_options) as writer:
        for _count, flat in flattener.flatten(releases):
            for name, rows in flat.items():
                for row in rows:
                    writer.writerow(name, row)

    for name in test_arrays:
        path = workdir / f"{name}.csv"
        assert not path.is_file()


def test_less_five_arrays_xlsx(spec_analyzed, releases, flatten_options, tmpdir):
    test_arrays = ["tenders_items", "tenders_items_addit", "tenders_tende"]
    flattener = Flattener(flatten_options, spec_analyzed.tables)
    tables = prepare_tables(spec_analyzed, flatten_options)
    workdir = Path(tmpdir)
    with XlsxWriter(workdir, tables, flatten_options) as writer:
        for _count, flat in flattener.flatten(releases):
            for name, rows in flat.items():
                for row in rows:
                    writer.writerow(name, row)

    path = workdir / "result.xlsx"
    xlsx_reader = openpyxl.load_workbook(path)
    for name in test_arrays:
        assert name not in xlsx_reader


def test_xlsx_only_no_default_columns(spec_analyzed, releases, tmpdir):
    flatten_options = FlattenOptions(**{"selection": {"tenders": {"split": True, "only": ["/tender/id"]}}})
    flattener = Flattener(flatten_options, spec_analyzed.tables)
    tables = prepare_tables(spec_analyzed, flatten_options)
    workdir = Path(tmpdir)
    with XlsxWriter(workdir, tables, flatten_options) as writer:
        for _count, flat in flattener.flatten(releases):
            for name, rows in flat.items():
                for row in rows:
                    writer.writerow(name, row)

    path = workdir / "result.xlsx"
    xlsx_reader = openpyxl.load_workbook(path)
    column = []
    for row in xlsx_reader["tenders"].rows:
        column.append(row[0].value)
    assert column[0] == "/tender/id"
    assert xlsx_reader["tenders"].max_column == 1
